# qpass: Frontend for pass (the standard unix password manager).
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2017
# URL: https://github.com/xolox/python-qpass

"""
Usage: qpass [OPTIONS] KEYWORD..

Search your password store for the given keywords or patterns and copy the
password of the matching entry to the clipboard. When more than one entry
matches you will be prompted to select the password to copy.

If you provide more than one KEYWORD all of the given keywords must match,
in other words you're performing an AND search instead of an OR search.

Instead of matching on keywords you can also enter just a few of the characters
in the name of a password, as long as those characters are in the right order.
Some examples to make this more concrete:

- The pattern 'pe/zbx' will match the name 'Personal/Zabbix'.
- The pattern 'ba/cc' will match the name 'Bank accounts/Creditcard'.

Supported options:

  -e, --edit

    Edit the matching entry instead of copying it to the clipboard.

  -l, --list

    List the matching entries on standard output.

  -n, --no-clipboard

    Don't copy the password of the matching entry to the clipboard, instead
    show the password on the terminal (by default the password is copied to
    the clipboard but not shown on the terminal).

  -p, --password-store=DIRECTORY

    Search the password store in DIRECTORY. If this option isn't given
    the password store is located using the $PASSWORD_STORE_DIR
    environment variable. If that environment variable isn't
    set the directory ~/.password-store is used.

  -v, --verbose

    Increase logging verbosity (can be repeated).

  -q, --quiet

    Decrease logging verbosity (can be repeated).

  -h, --help

    Show this message and exit.
"""

# Standard library modules.
import getopt
import logging
import sys

# External dependencies.
import coloredlogs
from humanfriendly.terminal import output, usage, warning

# Modules included in our package.
from qpass import PasswordStore
from qpass.exceptions import PasswordStoreError

# Public identifiers that require documentation.
__all__ = (
    'edit_matching_entry',
    'list_matching_entries',
    'logger',
    'main',
    'show_matching_entry',
)

# Initialize a logger for this module.
logger = logging.getLogger(__name__)


def main():
    """Command line interface for the ``qpass`` program."""
    # Initialize logging to the terminal.
    coloredlogs.install()
    # Parse the command line arguments.
    program_opts = {}
    action = show_matching_entry
    try:
        options, arguments = getopt.gnu_getopt(sys.argv[1:], 'elnp:vqh', [
            'edit', 'list', 'no-clipboard', 'password-store=',
            'verbose', 'quiet', 'help',
        ])
        for option, value in options:
            if option in ('-e', '--edit'):
                action = edit_matching_entry
            elif option in ('-l', '--list'):
                action = list_matching_entries
            elif option in ('-n', '--no-clipboard'):
                program_opts['clipboard_enabled'] = False
            elif option in ('-p', '--password-store'):
                program_opts['directory'] = value
            elif option in ('-v', '--verbose'):
                coloredlogs.increase_verbosity()
            elif option in ('-q', '--quiet'):
                coloredlogs.decrease_verbosity()
            elif option in ('-h', '--help'):
                usage(__doc__)
                return
            else:
                raise Exception("Unhandled option! (programming error)")
        if not (arguments or action == list_matching_entries):
            usage(__doc__)
            return
    except Exception as e:
        warning("Error: %s", e)
        sys.exit(1)
    # Execute the requested action.
    try:
        program = PasswordStore(**program_opts)
        action(program, arguments)
    except PasswordStoreError as e:
        # Known issues don't get a traceback.
        logger.error("%s", e)
        sys.exit(1)
    except KeyboardInterrupt:
        # If the user interrupted an interactive prompt they most likely did so
        # intentionally, so there's no point in generating more output here.
        sys.exit(1)


def edit_matching_entry(program, arguments):
    """Edit the matching entry."""
    name = program.select_entry(*arguments)
    program.context.execute('pass', 'edit', name)


def list_matching_entries(program, arguments):
    """List the entries matching the given keywords/patterns."""
    output('\n'.join(program.smart_search(*arguments)))


def show_matching_entry(program, arguments):
    """Show the matching entry on the terminal (and copy the password to the clipboard)."""
    name = program.select_entry(*arguments)
    output(program.format_entry(name))
    program.copy_password(name)



# Test suite for the `qpass' Python package.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2017
# URL: https://github.com/xolox/python-qpass

"""Test suite for the `qpass` package."""

# Standard library modules.
import logging
import os
import platform

# External dependencies.
from humanfriendly.testing import (
    CaptureOutput,
    MockedProgram,
    PatchedAttribute,
    PatchedItem,
    TemporaryDirectory,
    TestCase,
    random_string,
    run_cli,
    touch,
)
from humanfriendly.terminal import ansi_strip
from humanfriendly.text import dedent
from mock import MagicMock

# The module we're testing.
from qpass import DIRECTORY_VARIABLE, PasswordStore, cli
from qpass.cli import main
from qpass.exceptions import (
    EmptyPasswordStoreError,
    MissingPasswordStoreError,
    NoMatchingPasswordError,
)

# Initialize a logger for this module.
logger = logging.getLogger(__name__)


class PasswordStoreTestCase(TestCase):

    """:mod:`unittest` compatible container for `qpass` tests."""

    def test_cli_invalid_option(self):
        """Test error handling of invalid command line options."""
        returncode, output = run_cli(main, '-x', merged=True)
        assert returncode != 0
        assert "Error:" in output

    def test_cli_list(self):
        """Test the output of ``qpass --list``."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'foo.gpg'))
            touch(os.path.join(directory, 'foo/bar.gpg'))
            touch(os.path.join(directory, 'Also with spaces.gpg'))
            returncode, output = run_cli(main, '--password-store=%s' % directory, '--list')
            assert returncode == 0
            entries = output.splitlines()
            assert 'foo' in entries
            assert 'foo/bar' in entries
            assert 'Also with spaces' in entries

    def test_cli_usage(self):
        """Test the command line usage message."""
        for options in [], ['-h'], ['--help']:
            returncode, output = run_cli(main, *options)
            assert "Usage:" in output

    def test_clipboard_enabled(self):
        """Test the detection whether the clipboard should be used."""
        # Make sure the clipboard is enabled by default on macOS.
        if platform.system().lower() == 'darwin':
            assert PasswordStore().clipboard_enabled is True
        else:
            # Make sure the clipboard is used when $DISPLAY is set.
            with PatchedItem(os.environ, 'DISPLAY', ':0'):
                assert PasswordStore().clipboard_enabled is True
            # Make sure the clipboard is not used when $DISPLAY isn't set.
            environment = os.environ.copy()
            environment.pop('DISPLAY', None)
            with PatchedAttribute(os, 'environ', environment):
                assert PasswordStore().clipboard_enabled is False

    def test_directory_variable(self):
        """Test support for ``$PASSWORD_STORE_DIR``."""
        with TemporaryDirectory() as directory:
            with PatchedItem(os.environ, DIRECTORY_VARIABLE, directory):
                program = PasswordStore()
                assert program.directory == directory

    def test_edit_entry(self):
        """Test editing of an entry on the command line."""
        # Create a fake password store that we can test against.
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'Personal', 'Zabbix.gpg'))
            touch(os.path.join(directory, 'Work', 'Zabbix.gpg'))
            # Make sure we're not running the real `pass' program because its
            # intended purpose is user interaction, which has no place in an
            # automated test suite :-).
            with MockedProgram('pass'):
                returncode, output = run_cli(
                    main,
                    '--password-store=%s' % directory,
                    '--edit',
                    'p/z',
                    merged=True,
                )
                assert returncode == 0
                assert 'Matched one entry: Personal/Zabbix' in output

    def test_empty_password_store_error(self):
        """Test the EmptyPasswordStoreError exception."""
        with TemporaryDirectory() as directory:
            program = PasswordStore(directory=directory)
            self.assertRaises(EmptyPasswordStoreError, program.smart_search)

    def test_format_entry(self):
        """Test human friendly formatting of password store entries."""
        with TemporaryDirectory() as directory:
            name = 'some/random/password'
            value = random_string()
            program = PasswordStore(clipboard_enabled=False, directory=directory)
            program.get_text = MagicMock(return_value=value)
            self.assertEquals(
                # We enable ANSI escape sequences but strip them before we
                # compare the generated string. This may seem rather pointless
                # but it ensures that the relevant code paths are covered :-).
                dedent(ansi_strip(program.format_entry(name, use_colors=True))),
                dedent('''
                    some / random / password

                    Password: {value}
                ''', value=value))

    def test_fuzzy_search(self):
        """Test fuzzy searching."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'Personal/Zabbix.gpg'))
            touch(os.path.join(directory, 'Work/Zabbix.gpg'))
            touch(os.path.join(directory, 'Something else.gpg'))
            program = PasswordStore(directory=directory)
            # Test a fuzzy search with multiple matches.
            assert program.fuzzy_search('zbx') == ['Personal/Zabbix', 'Work/Zabbix']
            # Test a fuzzy search with a single match.
            assert program.fuzzy_search('p/z') == ['Personal/Zabbix']
            # Test a fuzzy search with `the other' match.
            assert program.fuzzy_search('w/z') == ['Work/Zabbix']

    def test_get_password(self):
        """Test getting a password from an entry."""
        name = 'some/random/password'
        value = random_string()
        text = '\n'.join([value, '', 'This is the description'])
        program = PasswordStore()
        program.get_text = MagicMock(return_value=text)
        self.assertEquals(value, program.get_password(name))

    def test_missing_password_store_error(self):
        """Test the MissingPasswordStoreError exception."""
        with TemporaryDirectory() as directory:
            missing = os.path.join(directory, 'missing')
            program = PasswordStore(directory=missing)
            self.assertRaises(MissingPasswordStoreError, program.ensure_directory_exists)

    def test_no_matching_password_error(self):
        """Test the NoMatchingPasswordError exception."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'Whatever.gpg'))
            program = PasswordStore(directory=directory)
            self.assertRaises(NoMatchingPasswordError, program.smart_search, 'x')

    def test_password_discovery(self):
        """Test password discovery."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'foo.gpg'))
            touch(os.path.join(directory, 'foo/bar.gpg'))
            touch(os.path.join(directory, 'foo/bar/baz.gpg'))
            touch(os.path.join(directory, 'Also with spaces.gpg'))
            program = PasswordStore(directory=directory)
            assert len(program.entries) == 4
            assert program.entries[0] == 'Also with spaces'
            assert program.entries[1] == 'foo'
            assert program.entries[2] == 'foo/bar'
            assert program.entries[3] == 'foo/bar/baz'

    def test_select_entry(self):
        """Test password selection."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'foo.gpg'))
            touch(os.path.join(directory, 'bar.gpg'))
            touch(os.path.join(directory, 'baz.gpg'))
            program = PasswordStore(directory=directory)
            # Substring search.
            assert program.select_entry('fo') == 'foo'
            # Fuzzy search.
            assert program.select_entry('bz') == 'baz'

    def test_select_entry_interactive(self):
        """Test interactive password selection."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'foo.gpg'))
            touch(os.path.join(directory, 'bar.gpg'))
            touch(os.path.join(directory, 'baz.gpg'))
            # Select entries using the command line filter 'a' and then use
            # interactive selection to narrow the choice down to 'baz' by
            # specifying the unique substring 'z'.
            program = PasswordStore(directory=directory)
            with CaptureOutput(input='z'):
                assert program.select_entry('a') == 'baz'

    def test_show_entry(self):
        """Test showing of an entry on the terminal."""
        password = random_string()
        # Some voodoo to mock methods in classes that
        # have yet to be instantiated follows :-).
        mocked_class = type(
            'TestPasswordStore',
            (PasswordStore,),
            dict(get_text=MagicMock(return_value=password)),
        )
        with PatchedAttribute(cli, 'PasswordStore', mocked_class):
            with TemporaryDirectory() as directory:
                name = 'some/random/password'
                touch(os.path.join(directory, '%s.gpg' % name))
                returncode, output = run_cli(
                    main,
                    '--password-store=%s' % directory,
                    '--no-clipboard',
                    name,
                )
                assert returncode == 0
                assert dedent(output) == dedent(
                    """
                    {title}

                    Password: {password}
                    """,
                    title=name.replace('/', ' / '),
                    password=password,
                )

    def test_simple_search(self):
        """Test simple substring searching."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'foo.gpg'))
            touch(os.path.join(directory, 'bar.gpg'))
            touch(os.path.join(directory, 'baz.gpg'))
            program = PasswordStore(directory=directory)
            assert program.simple_search('fo') == ['foo']
            assert program.simple_search('a') == ['bar', 'baz']
            assert program.simple_search('b', 'z') == ['baz']

    def test_smart_search(self):
        """Test smart searching."""
        with TemporaryDirectory() as directory:
            touch(os.path.join(directory, 'abcdef.gpg'))
            touch(os.path.join(directory, 'aabbccddeeff.gpg'))
            touch(os.path.join(directory, 'Google.gpg'))
            program = PasswordStore(directory=directory)
            # Test a substring match that avoids fuzzy matching.
            assert program.smart_search('abc') == ['abcdef']
            # Test a fuzzy match to confirm that the fall back works.
            assert program.smart_search('gg') == ['Google']



# qpass: Frontend for pass (the standard unix password manager).
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2017
# URL: https://github.com/xolox/python-qpass

"""Custom exceptions raised by :mod:`qpass`."""

# Public identifiers that require documentation.
__all__ = (
    'EmptyPasswordStoreError',
    'MissingPasswordStoreError',
    'NoMatchingPasswordError',
    'PasswordStoreError',
)


class PasswordStoreError(Exception):

    """Base class for custom exceptions raised by :mod:`qpass`."""


class MissingPasswordStoreError(PasswordStoreError):

    """Raised when the password store directory doesn't exist."""


class EmptyPasswordStoreError(PasswordStoreError):

    """Raised when the password store is empty."""


class NoMatchingPasswordError(PasswordStoreError):

    """Raised when no matching password can be selected."""



# qpass: Frontend for pass (the standard unix password manager).
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2017
# URL: https://github.com/xolox/python-qpass

"""
Frontend for pass_, the standard unix password manager.

.. _pass: https://www.passwordstore.org/
"""

# Standard library modules.
import logging
import os
import platform
import re

# External dependencies.
from executor import execute
from executor.contexts import LocalContext
from humanfriendly import Timer, format_path, parse_path
from humanfriendly.terminal import HIGHLIGHT_COLOR, ansi_wrap, terminal_supports_colors
from humanfriendly.prompts import prompt_for_choice
from humanfriendly.text import concatenate, format, pluralize, split, trim_empty_lines
from natsort import natsort
from proc.gpg import get_gpg_variables
from property_manager import PropertyManager, cached_property, clear_property, mutable_property, set_property
from verboselogs import VerboseLogger

# Modules included in our package.
from qpass.exceptions import (
    EmptyPasswordStoreError,
    MissingPasswordStoreError,
    NoMatchingPasswordError,
)

# Public identifiers that require documentation.
__all__ = (
    'DEFAULT_DIRECTORY',
    'DIRECTORY_VARIABLE',
    'PasswordStore',
    '__version__',
    'create_fuzzy_pattern',
    'logger',
)

# Semi-standard module versioning.
__version__ = '1.0.1'

DEFAULT_DIRECTORY = '~/.password-store'
"""
The default password storage directory (a string).

The value of :data:`DEFAULT_DIRECTORY` is normalized using
:func:`~humanfriendly.parse_path()`.
"""

DIRECTORY_VARIABLE = 'PASSWORD_STORE_DIR'
"""The environment variable that sets the password storage directory (a string)."""

KEY_VALUE_PATTERN = re.compile(r'^(.+\S):\s+(\S.*)$')
"""A compiled regular expression to recognize "Key: Value" lines."""

# Initialize a logger for this module.
logger = VerboseLogger(__name__)


class PasswordStore(PropertyManager):

    """Python API for the `qpass` program."""

    @mutable_property(cached=True)
    def clipboard_enabled(self):
        """
        :data:`True` to use the clipboard, :data:`False` otherwise.

        This property is :data:`True` on macOS (when the output of
        :func:`platform.system()` is 'Darwin') or when the environment variable
        ``$DISPLAY`` is set, otherwise it defaults to :data:`False`.
        """
        return platform.system().lower() == 'darwin' or bool(os.environ.get('DISPLAY'))

    @mutable_property(cached=True)
    def context(self):
        """
        An execution context created using :mod:`executor.contexts`.

        The value of :attr:`context` defaults to a
        :class:`~executor.contexts.LocalContext` object with the following
        characteristics:

        - The working directory of the execution context is set to the
          value of :attr:`directory`.

        - The environment variable given by :data:`DIRECTORY_VARIABLE` is set
          to the value of :attr:`directory`.

        :raises: :exc:`.MissingPasswordStoreError` when :attr:`directory`
                 doesn't exist.
        """
        # Make sure the directory exists.
        self.ensure_directory_exists()
        # Prepare the environment variables.
        environment = {DIRECTORY_VARIABLE: self.directory}
        try:
            # Try to enable the GPG agent in headless sessions.
            environment.update(get_gpg_variables())
        except Exception:
            # If we failed then let's at least make sure that the
            # $GPG_TTY environment variable is set correctly.
            environment.update(GPG_TTY=execute('tty', capture=True, check=False, tty=True, silent=True))
        return LocalContext(directory=self.directory, environment=environment)

    @mutable_property(cached=True)
    def directory(self):
        """
        The pathname of the password storage directory (a string).

        When the environment variable given by :data:`DIRECTORY_VARIABLE` is
        set the value of that environment variable is used, otherwise
        :data:`DEFAULT_DIRECTORY` is used. In either case the resulting
        directory pathname is normalized using
        :func:`~humanfriendly.parse_path()`.

        When you set the :attr:`directory` property, the value you set will be
        normalized using :func:`~humanfriendly.parse_path()` and the computed
        value of the :attr:`context` property is cleared.
        """
        return parse_path(os.environ.get(DIRECTORY_VARIABLE, DEFAULT_DIRECTORY))

    @directory.setter
    def directory(self, value):
        """Normalize the value of :attr:`directory` when it's set."""
        # Normalize the value of `directory'.
        set_property(self, 'directory', parse_path(value))
        # Clear the computed values of `context' and `entries'.
        clear_property(self, 'context')
        clear_property(self, 'entries')

    @cached_property
    def entries(self):
        """The names of the passwords stored in :attr:`directory` (a list of strings)."""
        timer = Timer()
        passwords = []
        logger.info("Scanning %s ..", format_path(self.directory))
        listing = self.context.capture('find', '-type', 'f', '-name', '*.gpg', '-print0')
        for filename in split(listing, '\0'):
            basename, extension = os.path.splitext(filename)
            if extension == '.gpg':
                # We use os.path.normpath() to remove the leading `./' prefixes
                # that `find' adds because it searches the working directory.
                passwords.append(os.path.normpath(basename))
        logger.verbose("Found %s in %s.", pluralize(len(passwords), "password"), timer)
        return natsort(passwords)

    def copy_password(self, name):
        """
        Copy a password to the clipboard.

        :param name: The name of the entry (a string).

        The password is only copied when :attr:`clipboard_enabled` is
        :data:`True`.
        """
        if self.clipboard_enabled:
            self.context.execute('pass', 'show', '--clip', name)

    def ensure_directory_exists(self):
        """
        Make sure :attr:`directory` exists.

        :raises: :exc:`.MissingPasswordStoreError` when the password storage
                 directory doesn't exist.
        """
        if not os.path.isdir(self.directory):
            msg = "The password storage directory doesn't exist! (%s)"
            raise MissingPasswordStoreError(msg % self.directory)

    def get_text(self, name):
        """
        Get the text of an entry in the password store.

        :param name: The name of the entry (a string).
        :returns: The entry's text (a string).
        """
        return self.context.capture('pass', 'show', name)

    def get_password(self, name):
        """
        Get the password of an entry in the password store.

        :param name: The name of the entry (a string).
        :returns: The entry's password (a string).
        """
        text = self.get_text(name)
        lines = text.splitlines()
        return lines[0]

    def format_entry(self, name, use_colors=None):
        """
        Format an entry's text for viewing on a terminal.

        :param name: The name of the entry (a string).
        :param use_colors: :data:`True` to use ANSI escape sequences,
                           :data:`False` otherwise. When this is :data:`None`
                           :func:`~humanfriendly.terminal.terminal_supports_colors()`
                           will be used to detect whether ANSI escape sequences
                           are supported.
        :returns: The formatted entry (a string).
        """
        # Determine whether we can use ANSI escape sequences.
        if use_colors is None:
            use_colors = terminal_supports_colors()
        # Get the entry's text.
        entry = self.get_text(name)
        # Extract the password (first line) from the entry.
        lines = entry.splitlines()
        password = lines.pop(0).strip()
        entry = trim_empty_lines('\n'.join(lines))
        # Show the password only when the clipboard is disabled.
        if not self.clipboard_enabled:
            entry = "Password: %s\n%s" % (password, entry)
        # Add the name to the entry (only when there's something to show).
        if entry and not entry.isspace():
            title = ' / '.join(split(name, '/'))
            if use_colors:
                title = ansi_wrap(title, bold=True)
            entry = "%s\n\n%s" % (title, entry)
        # Highlight the entry's text using ANSI escape sequences.
        lines = []
        for line in entry.splitlines():
            # Check for a "Key: Value" line.
            match = KEY_VALUE_PATTERN.match(line)
            if match:
                key = "%s:" % match.group(1).strip()
                value = match.group(2).strip()
                if use_colors:
                    # Highlight the key.
                    key = ansi_wrap(key, color=HIGHLIGHT_COLOR)
                    # Underline hyperlinks in the value.
                    tokens = value.split()
                    for i in range(len(tokens)):
                        if '://' in tokens[i]:
                            tokens[i] = ansi_wrap(tokens[i], underline=True)
                    # Replace the line with a highlighted version.
                    line = key + ' ' + ' '.join(tokens)
            lines.append('  ' + line)
        return '\n%s\n' % '\n'.join(lines)

    def fuzzy_search(self, *filters):
        """
        Perform a "fuzzy" search that matches the given characters in the given order.

        :param filters: The pattern(s) to search for.
        :returns: The matched password names (a list of strings).
        """
        matches = []
        logger.verbose("Performing fuzzy search on %s (%s) ..",
                       pluralize(len(filters), "pattern"),
                       concatenate(map(repr, filters)))
        patterns = list(map(create_fuzzy_pattern, filters))
        for passwd in self.entries:
            if all(p.search(passwd) for p in patterns):
                matches.append(passwd)
        logger.log(logging.INFO if matches else logging.VERBOSE,
                   "Matched %s using fuzzy search.", pluralize(len(matches), "password"))
        return matches

    def select_entry(self, *arguments):
        """
        Select a password from the available choices.

        :param arguments: Refer to :func:`smart_search()`.
        :returns: The name of a password (a string) or :data:`None`
                  (when no password matched the given `arguments`).
        """
        matches = self.smart_search(*arguments)
        if len(matches) > 1:
            logger.info("More than one match, prompting for choice ..")
            return prompt_for_choice(matches)
        else:
            logger.info("Matched one entry: %s", matches[0])
            return matches[0]

    def simple_search(self, *keywords):
        """
        Perform a simple search for case insensitive substring matches.

        :param keywords: The string(s) to search for.
        :returns: The matched password names (a generator of strings).

        Only passwords whose names matches *all*  of the given keywords are
        returned.
        """
        matches = []
        keywords = [kw.lower() for kw in keywords]
        logger.verbose("Performing simple search on %s (%s) ..",
                       pluralize(len(keywords), "keyword"),
                       concatenate(map(repr, keywords)))
        for passwd in self.entries:
            normalized = passwd.lower()
            if all(kw in normalized for kw in keywords):
                matches.append(passwd)
        logger.log(logging.INFO if matches else logging.VERBOSE,
                   "Matched %s using simple search.", pluralize(len(matches), "password"))
        return matches

    def smart_search(self, *arguments):
        """
        Perform a smart search on the given keywords or patterns.

        :param arguments: The keywords or patterns to search for.
        :returns: The matched password names (a list of strings).
        :raises: The following exceptions can be raised:

                 - :exc:`.NoMatchingPasswordError` when no matching passwords are found.
                 - :exc:`.EmptyPasswordStoreError` when the password store is empty.

        This method first tries :func:`simple_search()` and if that doesn't
        produce any matches it will fall back to :func:`fuzzy_search()`. If no
        matches are found an exception is raised (see above).
        """
        matches = self.simple_search(*arguments)
        if not matches:
            logger.verbose("Falling back from substring search to fuzzy search ..")
            matches = self.fuzzy_search(*arguments)
        if not matches:
            if len(self.entries) > 0:
                raise NoMatchingPasswordError(format(
                    "No passwords matched the given arguments! (%s)",
                    concatenate(map(repr, arguments)),
                ))
            else:
                raise EmptyPasswordStoreError(format(
                    "You don't have any passwords yet! (no *.gpg files found in %s)",
                    format_path(self.directory),
                ))
        return matches


def create_fuzzy_pattern(pattern):
    """
    Convert a string into a fuzzy regular expression pattern.

    :param pattern: The input pattern (a string).
    :returns: A compiled regular expression object.

    This function works by adding ``.*`` between each of the characters in the
    input pattern and compiling the resulting expression into a case
    insensitive regular expression.
    """
    return re.compile('.*'.join(map(re.escape, pattern)), re.IGNORECASE)
